#include "single_var.h"

using namespace single_var;
using namespace std;
using namespace Eigen;

int main( int argc, char * argv[] )
{

    const int NUM_ARGC = 2;
    if ( ( argc != NUM_ARGC ))
    {
        cout << "Usage: " <<
           argv[0] << " config_filename " << endl;
        return ARG_ERROR;
    }

    stringstream convert_config_filename(argv[1]);
    string config_filename;
    if (!(convert_config_filename >> config_filename))
    {
        cout << "Could not convert command line ";
        cout << "argument to string type for config_filename" << endl;
        return ARG_ERROR;
    }
    string geno_filename;
    string out_filename;
    size_t start_rep;
    size_t num_reps;
    double MAF_cutoff;
    bool use_pheno;
    string pheno_filename;
    size_t out_freq;
    double min_call_rate;
    const int rv_config_file = 
       read_configfile( config_filename, geno_filename, out_filename,
                        start_rep, num_reps, MAF_cutoff, 
                        use_pheno, pheno_filename, out_freq, min_call_rate );
    if ( NO_ERROR != rv_config_file )
    {
       cout << "Error reading config_file.  Exiting." << endl;
       return EXIT_FAILURE;
    }
    cout << endl;
    cout << "Reading configuration from " << config_filename << endl;
    cout << "Genomic data file : " << geno_filename << endl;
    cout << "Output file : " << out_filename << endl;
    cout << "start_rep : " << start_rep << endl;
    cout << "num_reps : " << num_reps << endl;
    cout << "MAF_cutoff : " << MAF_cutoff << endl;
    cout << "min_call_rate : " << min_call_rate << endl;
    if ( false == use_pheno )
    {
        cout << "No phenotype/covariate file specified." << endl;
    }
    else
    {
        cout << "Phenotype/Covariates will be read from " 
             << pheno_filename << endl; 
    }
    cout << "out_freq : " << out_freq << endl;
    cout << endl;

    std::vector< subject_data_type > subject_data_vec;
    MatrixXd Y_mat;
    MatrixXd X_mat;

    if ( use_pheno )
    {
       const int rv_pheno =
           read_pheno_data( pheno_filename, X_mat, Y_mat, subject_data_vec );

    
      if ( rv_pheno != NO_ERROR )
      {
        cout << "Error: read_pheno_data() call in main() " <<
                "returned " << rv_pheno << endl;
        cout << "exiting" << endl;
        return EXIT_FAILURE;
    }
  }


    std::vector< MatrixXd > geno_mat_vec;
    std::vector< string > geno_names_vec;
    const int rv_geno = read_geno_file( geno_filename, 
                        geno_mat_vec, geno_names_vec, subject_data_vec,
                        min_call_rate );
    if ( rv_geno != NO_ERROR )
    {
        cout << "read_geno() error : read_geno() returned " << rv_geno << endl;
        return FILE_ERROR;
    }

	//first remove all non-common snps from G matrices

    size_t count_snps_above_MAF_cutoff = 0;
	for ( auto & G_mat : geno_mat_vec )
    {
       const size_t NUM_COOR_G  = G_mat.rows();
        //VectorXd c_vec;
        //calc_Burden_c_vec( G_mat, MAF_cutoff, c_vec );
        //cout << "c_vec == " << endl << c_vec << endl;
        VectorXd min_MAF_c_vec;
        calc_above_min_MAF_c_vec( G_mat, MAF_cutoff, min_MAF_c_vec );
        size_t NUM_COMMON_GENO_COOR = 0;
        //cout << "min_MAF_c_vec == " << endl << min_MAF_c_vec << endl;
        for ( size_t index = 0; index < min_MAF_c_vec.rows(); ++index )
        {
          	const size_t my_int = ( min_MAF_c_vec( index ) > 0.5 ); 
            NUM_COMMON_GENO_COOR += my_int;
        }
        MatrixXd common_G_mat = MatrixXd::Zero( NUM_COMMON_GENO_COOR, G_mat.cols() );
        for ( size_t coor = 0, known_index = 0;
                        coor < G_mat.rows();
                        ++coor )
        {
            //int my_int = int( c_vec( coor ) );
			const bool my_bool = ( min_MAF_c_vec( coor ) > 0.5 );
            if ( ! my_bool )
            {
                //rare snp, skip
                continue;
            }
            common_G_mat.block( known_index, 0, 1, G_mat.cols() ) =
                G_mat.row( coor );
            known_index++;
        }
        //cout << "G_mat == " << endl << G_mat << endl;        
        G_mat = common_G_mat;
    	if ( G_mat.rows() )
		{
			count_snps_above_MAF_cutoff++;
		}
	} //loop through geno_mat_vec

    bool non_empty_G_mat = false;
    for ( const auto & G_mat : geno_mat_vec )
    {
        if ( G_mat.rows() > 0 )
        {
           non_empty_G_mat = true;
           break;
        }
    }
    if ( false == non_empty_G_mat )
    {
        cout << "No SNPs were left after MAF_cutoff filter." << endl;
        return EXIT_FAILURE;
    }

        for ( auto & G_mat : geno_mat_vec )
        {

          G_mat.transposeInPlace();
        }

    if ( use_pheno == false )
    {
        const int rv_resample = 
            resample_threshold_linear_nocovar(  
                                       geno_mat_vec, 
                                       out_filename, start_rep, num_reps, out_freq );
	    
        if ( rv_resample != NO_ERROR )
        {
            cout << "error : resample_threshold_linear_nocovar  returned " 
                 << rv_geno << endl;
            return FILE_ERROR;
        }
        cout << "Done." << endl;
        return NO_ERROR;
    }
    //USE_PHENO == true

    VectorXd Y_vec = static_cast<VectorXd>( Y_mat );
    //cout << "Y_vec.rows() == " << Y_vec.rows() << endl;
    //cout << "X_mat.rows() == " << X_mat.rows() << endl;        
    //cout << "X_mat.cols() == " << X_mat.cols() << endl;
    const int rv_resample = resample_Burden_threshold_linear(
            X_mat,
            Y_vec,
            geno_mat_vec,
            out_filename,
            start_rep,
            num_reps,
            out_freq ); 

    if ( rv_resample != NO_ERROR )
    {
        cout << "Error: resample_Burden_threshold_linear() call in main() " <<
                "returned " << rv_resample << endl;
        cout << "exiting" << endl;
        return EXIT_FAILURE;
    }
    
    cout << "Done." << endl;
    return NO_ERROR;
} //main
