#include "thresh_region_util.h"
#include <iostream>
#include <vector>
#include <map>
#include <eigen3/Eigen/Dense>
using namespace std;
using namespace Eigen;
using namespace thresh;

int run_simulation_without_pheno(
                   SIM_TYPE my_sim_type,
                   const string & out_filename,
                   const std::vector< MatrixXd > & geno_mat_vec,
                   size_t start_rep,
                   size_t num_reps,
                   size_t min_MAC,
                   size_t out_freq
                   )
{
    string full_out_filename;
    //cout << "run_simulation_without_pheno() : " << endl;
    cout << "Simulation without phenotype and without covariates." << endl;
    if (my_sim_type != SKAT_ONLY )
    {
        cout << "Beginning Burden simulation." << endl;
        full_out_filename = out_filename + ".BURDEN.txt";
        cout << "Resampling output file: " << full_out_filename << endl;
        const int threshold_rv = 
         resample_Burden_threshold_linear_nocovar(
            geno_mat_vec,
            full_out_filename,
            start_rep,
            num_reps,
            out_freq,
            min_MAC
            );
        if ( threshold_rv != NO_ERROR )
        {
             cout << "Error: resample_Burden_threshold_linear_nocovar returned non-zero " <<
                    "value == " << threshold_rv << endl;  
                return SIM_ERROR;
        }
       cout << "Burden simulation completed. " << endl;
       if ( my_sim_type == BURDEN_ONLY )
       {
           return NO_ERROR;
       }
    } //if (my_sim_type != SKAT_ONLY )
    //else my_sim_type is either SKAT_ONLY or BURDEN_AND_SKAT
    //run SKAT simulations without pheno
    full_out_filename = out_filename + ".SKAT.txt";
    cout << "Resampling output file: " << full_out_filename << endl;
    const int threshold_rv = 
       resample_SKAT_threshold_linear_nocovar(
         geno_mat_vec, 
         full_out_filename,
         start_rep,
         num_reps,
         out_freq,
         min_MAC );
    if ( threshold_rv != NO_ERROR )
    {
        cout << "Error: resample_SKAT_threshold_linear_nocovar returned " <<
                threshold_rv << endl;
        return SIM_ERROR;
    }     
   cout << "SKAT simulation completed. " << endl;
   return NO_ERROR;
} //***** run_simulation_without_pheno() method


int run_simulation_with_pheno( 
                   SIM_TYPE my_sim_type,
                   const MatrixXd & X_mat,
                   const MatrixXd & Y_mat,
                   const string & out_filename,
                   const std::vector< MatrixXd > & geno_mat_vec,
                   size_t start_rep,
                   size_t num_reps,
                   size_t min_mac,
                   size_t out_freq
                   )
{
    cout << "Simulation with phenotype and with covariates." << endl;
       VectorXd Y_vec = static_cast<VectorXd>( Y_mat );
       //cout << "Simulation using phenotype and covariates." << endl;
       string full_out_filename;
       if ( my_sim_type != SKAT_ONLY )
       {
           cout << "Reginning Burden simulation." << endl;
           full_out_filename = out_filename + ".BURDEN.txt";
           cout << "Resampling output file: " << full_out_filename << endl;
           const int threshold_rv = 
            resample_Burden_threshold_linear(
                                X_mat,
                                Y_vec,
                                geno_mat_vec,
                                full_out_filename,
                                start_rep,
                                num_reps,
                                out_freq, min_mac );    
           if ( threshold_rv != NO_ERROR )
           {
               cout << "Error: resample_Burden_threshold_linear returned non-zero " <<
                    "value == " << threshold_rv << endl;
               return SIM_ERROR;				
           }
           cout << "Burden simulation completed. " << endl;
           if ( my_sim_type == BURDEN_ONLY )
           {
               return NO_ERROR;
           }
       } // if ( my_sim_type != SKAT_ONLY )
       //else my_sim_type is either SKAT_ONLY or BURDEN_AND_SKAT

       //run SKAT simulations with pheno
       full_out_filename = out_filename + ".SKAT.txt";
       cout << "Resampling output file: " << full_out_filename << endl;
       const int threshold_rv = 
           resample_SKAT_threshold_linear(
              X_mat,
              Y_vec,
              geno_mat_vec,
              full_out_filename,
              start_rep,
              num_reps, out_freq, min_mac );
        if ( threshold_rv != NO_ERROR )
        {
            cout << "Error: resample_SKAT_threshold_linear() method " <<
                    "returned " << threshold_rv << endl;
            return SIM_ERROR;
        }       

     return NO_ERROR;
} //run_simulation_with_pheno() method


//**********************************
//**********  main *****************
//**********************************
int main( int argc, char * argv[] )
{
    const double min_MAF = 1e-100;
    const bool remove_too_rare = false;
    const bool TEST_WINDOW_MAP = false;
    if ( TEST_WINDOW_MAP == true )
    {
       int map_test_rv = make_window_map_from_geno( 
           "geno.txt", "geno.txt.map.cpp_code.2.2.txt",
           2, 2 );
       if ( map_test_rv != NO_ERROR )
       {
          cout << "First call to make_window_map_from_geno() failed." << endl;
          return -1;
       }


       map_test_rv = make_window_map_from_geno(           
         "geno.txt", "geno.txt.map.cpp_code.2.1.txt",
         2, 1 );
       if ( map_test_rv != NO_ERROR )
       {
          cout << "Second call to make_window_map_from_geno() failed." << endl;
          return -2;
       }
       cout << "Done with make_window_map_from_geno() tests." << endl;
       return 0;
    } // if TEST_WINDOW_MAP == true
//***end test map code
//*************************************************************
 
//***********************************************************
//*** Variables in this ection are important for both gene
//*** and window-based simulations.
//***
    string geno_filename;
    string out_filename;    
    size_t start_rep;
    size_t num_reps;
    string pheno_filename;
    double min_call_rate;
    const int NUM_ARGC = 2;
    if ( ( argc != NUM_ARGC ))
    {
        cout << "Usage: " <<
           argv[0] << " config_filename " << endl;
        return ARG_ERROR;
    }
    stringstream convert_config_filename(argv[1]);
    string config_filename;
    if (!(convert_config_filename >> config_filename))
    {
        cout << "Could not convert command line ";
        cout << "argument to string type for config_filename" << endl;
        return ARG_ERROR;
    }
    bool use_pheno;
    SIM_TYPE sim_type;
    size_t window_size;
    size_t window_step;
    string map_filename;
    bool make_map_from_window;
    double MAF_cutoff;
    size_t min_MAC;
    size_t out_freq;
    const int rv_config_file = 
      read_configfile(
                 config_filename,
                 geno_filename,
                 out_filename,
                 start_rep,
                 num_reps,
                 MAF_cutoff,
                 min_MAC,
                 use_pheno,
                 pheno_filename,
                 sim_type,
                 window_size,
                 window_step,
                 make_map_from_window,
                 map_filename,
                 out_freq,
                 min_call_rate  );
    if ( NO_ERROR != rv_config_file )
    {
       cout << "Error reading from configuration file." << endl;
       return EXIT_FAILURE;
    }
    if ( NO_ERROR != rv_config_file )
    {
       cout << "Error reading config_file.  Exiting." << endl;
       return EXIT_FAILURE;
    }
    cout << endl;
    cout << "Reading configuration from " << config_filename << endl;
    cout << "Genomic data file : " << geno_filename << endl;
    cout << "Output file prefix : " << out_filename << endl;
    cout << "start_rep : " << start_rep << endl;
    cout << "num_reps : " << num_reps << endl;
    cout << "MAF_cutoff : " << MAF_cutoff << endl;
    cout << "min_MAC : " << min_MAC << endl;
    cout << "min_call_rate : " << min_call_rate << endl;
    if ( false == use_pheno )
    {
        cout << "No phenotype/covariate file specified." << endl;
    }
    else
    {
        cout << "Phenotype/Covariates will be read from "
             << pheno_filename << endl;
    }
    cout << "out_freq : " << out_freq << endl;

    if ( true == make_map_from_window )
    {
       cout << "window_size : " << window_size << endl;
       cout << "window_step : " << window_step << endl;
       map_filename = geno_filename + ".window." + to_string( window_size ) +
            ".step." + to_string( window_step ) + ".map.txt";
       const int rv_make_window_map = 
         make_window_map_from_geno( geno_filename, map_filename, window_size, window_step );
       if ( NO_ERROR != rv_make_window_map ) 
       {

          cout << "Error: Could not generate map file. "
             << "rv_make_window_map == " << rv_make_window_map << endl;
          return EXIT_FAILURE;
        } 
       cout << "Wrote map data based on window data to " << map_filename << endl;
    } //if

    cout << "map_filename : " << map_filename << endl;
    cout << "start_rep : " << start_rep << endl;
    cout << "num_reps : " << num_reps << endl;
    cout << "sim_type : ";
    if ( sim_type == BURDEN_ONLY )
    {
        cout << "Burden" << endl;
    }
    else
    if ( sim_type == SKAT_ONLY )
    {
        cout << "SKAT" << endl;
    }
    else
    if ( sim_type == BURDEN_AND_SKAT )
    {
        cout << "Both Burden and SKAT." << endl;
    }
    else
    {
        cout << "error : Invalid sim_type " << sim_type << endl;
        cout << "Exiting." << endl;
        return EXIT_FAILURE;
    }
    int rv = FILE_ERROR;
    //mapping file
    map<string, vector<string> > snp_to_gene_map;
    
    rv = read_map_file(
            map_filename, snp_to_gene_map);
    if ( rv != NO_ERROR )
    {
       cout << "Error: read_map_file() call in main() " 
            << "returned " << rv << endl;
       cout << "exiting." << endl;
       return EXIT_FAILURE;
    }


    //geno file
    std::vector< MatrixXd > geno_mat_vec;
    std::vector< string > gene_names_vec;
    std::vector< subject_data_type > subject_data_vec_missing_pheno;
    
    MatrixXd X_mat;
    MatrixXd Y_mat;
    if (use_pheno)
    {
       const int rv_pheno =             
           read_pheno_data( pheno_filename, X_mat, Y_mat, 
                    subject_data_vec_missing_pheno );


        if ( rv_pheno != NO_ERROR )
        {
            cout << "Error: read_pheno_data() call in main() " <<
                "returned " << rv_pheno << endl;
            cout << "exiting" << endl;
            return EXIT_FAILURE;
        }
     }


    rv = read_geno_file( geno_filename, snp_to_gene_map,  
                        geno_mat_vec, 
                         gene_names_vec,
                        subject_data_vec_missing_pheno, min_call_rate );
    if ( rv != NO_ERROR )
    {
        cout << "Error : read_geno_file() call in main() returned " 
             << rv << endl;
        cout << "exiting." << endl;
        return EXIT_FAILURE;
    }


    //cout << "MAF_cutoff == " << MAF_cutoff << endl;
    //first remove all non-rare snps from G matrices
    for ( auto & G_mat : geno_mat_vec )
    {
       const size_t NUM_COOR_G  = G_mat.rows();
        VectorXd c_vec;
        calc_Burden_c_vec( G_mat, MAF_cutoff, c_vec );
        //cout << "c_vec == " << endl << c_vec << endl;
        VectorXd min_MAF_c_vec;
        calc_above_min_MAF_c_vec( G_mat, min_MAF, min_MAF_c_vec );
        size_t NUM_RARE_GENO_COOR = 0;
        for ( size_t index = 0; index < c_vec.rows(); ++index )
        {
            //int my_int = int( c_vec( index ) );
          	int my_int;
			if ( c_vec( index ) > 0.5 )                
            {
                if ( ! remove_too_rare )
                {
                    my_int = 1;   
                }
                else
                {
                    my_int = ( min_MAF_c_vec( index ) > 0.5 );
                }
            }
            else
            {
                my_int = 0;
            }
            NUM_RARE_GENO_COOR += my_int;
        }
        //MatrixXd rare_G_mat = G_mat;
        MatrixXd rare_G_mat = MatrixXd::Zero( NUM_RARE_GENO_COOR, G_mat.cols() );
        for ( size_t coor = 0, known_index = 0;
                        coor < G_mat.rows();
                        ++coor )
        {
            //int my_int = int( c_vec( coor ) );
			int my_int;
            if ( c_vec( coor ) > 0.5 )
            {
                if ( ! remove_too_rare )
                {
                    my_int = 1;
                }
                else
                {
                    my_int = ( min_MAF_c_vec( coor ) > 0.5 );
                }
            }
            else
            {
                my_int = 0;
            }
            if ( ! my_int )
            {
                //common snp, skip
                continue;
            }
            rare_G_mat.block( known_index, 0, 1, G_mat.cols() ) =
                G_mat.row( coor );
            known_index++;
        }
        
        G_mat = rare_G_mat;
    } //loop through geno_mat_vec
    std::cout << "removed common SNPs" << std::endl;
    bool non_empty_G_mat = false;
    for ( const auto & G_mat : geno_mat_vec )
    {
        if ( G_mat.rows() > 0 )
        {
           non_empty_G_mat = true;
           break;
        }
    }
    if ( false == non_empty_G_mat )
    {
        cout << "No genes were left after MAF_cutoff filter." << endl;
        return EXIT_FAILURE;
    }

    //******************
    //Simulations with phenotype and covariates
    //******************
    if ( use_pheno )
    {
       const int rv_sim = 
           run_simulation_with_pheno( 
             sim_type, X_mat, Y_mat, out_filename,
             geno_mat_vec, start_rep, num_reps, min_MAC, out_freq );       
       if ( rv_sim != NO_ERROR )
       {
          cout << "Error: run_simulation_with_pheno() returned "
               << rv_sim << endl;
          return EXIT_FAILURE;
       }
       
       return EXIT_SUCCESS;
    }  //USE_PHENO is true
    //***************
    //End of simulations with phenotype and covariates.
    //****************


    //*************************
    //Simulations without phenotype and without covariates.
    //************************
    //USE_PHENO is false
    const int rv_sim = run_simulation_without_pheno( 
           sim_type, out_filename, geno_mat_vec, start_rep, num_reps, min_MAC, out_freq );
    if ( rv_sim != NO_ERROR )
    {
        cout << "Error: run_simulation_without_pheno() returned "
               << rv_sim << endl;
        return EXIT_FAILURE;
    }
    return EXIT_SUCCESS;
    //***************
    //End of simulations without phenotype and without covariates.
    //****************

} // ******end main

